#include "xpinball.h"

#define SMALL 1.0e-15

void periodic(data)
     Data *data;
{
  Vector tmp_v, tmp_v1, A, B, C, start[3], Bt;
  char buf[50], orbit[100], bin_orbit[100], orbit_buf[250];
  int i, j, N, old, next;
  double len, old_len, tmp, tau, r, J[4], lam;
  double tmp1, tmp2, tmp3, tmp4, old_Bx, M, Mp, tmp5, tmp6;
  Display *dpy = XtDisplay(data->drawing_a);
  void binary_to_ternary();

  if (data->type=='b')
    binary_to_ternary(data);
  else
    ternary_to_binary(data);

  N = data->period;
  if (N==0)
    return;

  addvx(start[0], data->disk[0], data->a, 0);
  addvx(start[1], data->disk[1], -data->a, 0);
  addvx(start[2], data->disk[2], 0, -data->a);

  for(i=0; i<N; i++)
    copyv(data->PosPO[i], start[data->orbit[i]]);

  len =5.0*N;
  j = 0;
  do {
    old_len = len;
    len = 0.0;
    old = N-1;
    for(i=0; i<N; i++) {
      subv(A, data->PosPO[old], data->disk[data->orbit[i]]);
      subv(B, data->PosPO[i], data->disk[data->orbit[i]]);
      subv(C, data->PosPO[(i+1)%N], data->disk[data->orbit[i]]);
      old = i;
      tmp1 = dotv(A,A) + data->asq;
      tmp2 = dotv(C,C) + data->asq;
/*
      do {
	j++;
	old_Bx = B.x;
	tmp3 = 1.0/sqrt(tmp1 - 2.0*dotv(A, B));
	tmp4 = 1.0/sqrt(tmp2 - 2.0*dotv(C, B));
	linear(B, tmp3, A, tmp4, C);
	tmp = data->a/hypotv(B);
	smulv(B, tmp, B);
      } while (fabs(B.x-old_Bx)>SMALL);
*/
      do {
	j++;
	old_Bx = B.x;
	tmp3 = 1.0/sqrt(tmp1 - 2.0*dotv(A, B));
	tmp4 = 1.0/sqrt(tmp2 - 2.0*dotv(C, B));
	setv(Bt, -B.y, B.x);
	tmp5 = dotv(A,Bt);
	tmp6 = dotv(C,Bt);
	M = tmp5*tmp3 + tmp6*tmp4;
	Mp = tmp5*tmp5*tmp3*tmp3*tmp3;
	Mp += tmp6*tmp6*tmp4*tmp4*tmp4;
	Mp -= tmp3*dotv(A,B) + tmp4*dotv(C,B);
	tmp = -M/Mp;
	linear(B, 1.0, B, tmp, Bt);
	tmp = data->a/hypotv(B);
	smulv(B, tmp, B);
      } while (fabs(M)>SMALL);
      addv(data->PosPO[i], B, data->disk[data->orbit[i]]);
      subv(tmp_v, A, B);
      len += hypotv(tmp_v);
      subv(tmp_v, C, B);
      len += hypotv(tmp_v);
    }
  } while(fabs(old_len-len) > SMALL);

  copyv(data->PosPO[N], data->PosPO[0]);

  if(data->newPO)
    XDrawSegments(dpy, XtWindow(data->drawing_a), data->gc_old_orbit,
		  data->xraysPO, data->no_raysPO);
  else
    data->newPO++;

  for (i=0; i<N; i++) {
    data->xraysPO[i].x1 = (short) (data->mag*data->PosPO[i].x + data->offset.x);
    data->xraysPO[i].y1 = (short) (-data->mag*data->PosPO[i].y + data->offset.y);
    data->xraysPO[i].x2 = (short) (data->mag*data->PosPO[i+1].x + data->offset.x);
    data->xraysPO[i].y2 = (short) (-data->mag*data->PosPO[i+1].y + data->offset.y);
  }
  data->no_raysPO = N;
  
  XDrawSegments(dpy, XtWindow(data->drawing_a),
		data->gc_orbit, data->xraysPO, data->no_raysPO);

  for(i=0; i<data->period; i++) {
    orbit[i] = data->orbit[i] + '1';
  }
  orbit[data->period] = '\0';
  for(i=0; i<data->bin_period; i++) {
    bin_orbit[i] = data->bin_orbit[i] + '0';
  }
  bin_orbit[data->bin_period] = '\0';
  sprintf(orbit_buf, "%s = B%s", orbit, bin_orbit);

  len = 0.0;
  if (data->type=='b')
    N = data->bin_period;
  J[0] = J[3] = 1.0;
  J[1] = J[2] = 0.0;
  for(i=0; i<data->bin_period; i++) {
    next = (i+1)%data->period;
    subv(tmp_v, data->PosPO[next], data->PosPO[i]);
    len += tau = hypotv(tmp_v);
    subv(tmp_v1, data->PosPO[i], data->disk[data->orbit[i]]);
    r = 2.0*data->a*tau/dotv(tmp_v, tmp_v1);
    tau *= data->ia;
    J[1] += J[0]*r;
    J[3] += J[2]*r;
    J[0] += J[1]*tau;
    J[2] += J[3]*tau;
  }

  j = 1;
  for(i=0; i<data->bin_period; i++) {
    j *= (data->bin_orbit[i])? -1:1;
  }

  tmp = j*(J[0]+J[3]);

  if (tmp>0.0)
    lam = 0.5*(tmp+sqrt(tmp*tmp-4.0));
  else
    lam = 0.5*(tmp-sqrt(tmp*tmp-4.0));

  i = data->period/data->bin_period;
  sprintf(buf, "Time = %dx(%f)   Stability = (%g)^%d", i, data->ia*len, lam, i);

  WriteLabels(data, orbit_buf, buf);
}

static char other[3][3] = {{0, 2, 1}, {2, 1, 0}, {1, 0, 2}};

void binary_to_ternary(data)
     Data *data;
{
  int i, j, bp, diff;

  bp = data->bin_period;

  if (bp==1) {
    if (data->bin_orbit[0]==0) {
      data->period = 2;
      data->orbit[0] = 0;
      data->orbit[1] = 1;
    } else {
      data->period = 3;
      data->orbit[0] = 0;
      data->orbit[1] = 1;
      data->orbit[2] = 2;
    }
    return;
  }

  data->orbit[0] = 0;
  data->orbit[1] = 1;
  for(i=2; i<bp; i++) {
    data->orbit[i] = (data->bin_orbit[i])? 
      other[data->orbit[i-1]][data->orbit[i-2]]: data->orbit[i-2];
  }
  j = bp;
  do {
    diff = 0;
    for(i=0; i<bp; i++) {
      data->orbit[j] = (data->bin_orbit[i])?
	other[data->orbit[j-1]][data->orbit[j-2]]: data->orbit[j-2];
      if (data->orbit[j]!=data->orbit[i])
	diff++;
      j++;
    }
  } while (diff);
  data->period = j - bp;
}

ternary_to_binary(data)
     Data *data;
{
  int i;
  int unique();

  if (data->period<2)
    return;

  data->bin_orbit[0] = (data->orbit[0]==data->orbit[data->period-2])? 0:1;
  data->bin_orbit[1] = (data->orbit[1]==data->orbit[data->period-1])? 0:1;

  for(i=2; i<data->period; i++) {
    data->bin_orbit[i] = (data->orbit[i]==data->orbit[i-2])? 0:1;
  }
  data->bin_orbit[data->period] = '\0';
  data->bin_period = unique(data->bin_orbit, data->period);
}

int unique(string, N)
     char *string;
     int N;
{
  int i, j, k, l;
  char *sub, first;
  void sort();

  sub = (char *) malloc(N);

  for(i=1; i<N; i++) {
    if (N%i)
      continue;
    for(l=0; l<i; l++) {
      for(j=0; j<i; j++)
	sub[j] = string[j];
      for(k=0; k<N; k+=i) {
	for(j=0; j<i; j++)
	  if (string[k+j]!=sub[j])
	    break;
	if(j!=i)
	  break;
      }
      if (k==N) {
	sort(sub, i);
	for(j=0; j<i; j++)
	  string[j] = sub[j];
	free(sub);
	return(i);
      }
      first = string[0];
      for(k=0; k<N-1; k++) {
	string[k] = string[k+1];
      }
      string[N-1] = first;
    }
  }
  sort(string, N);
  free(sub);
  return(N);
}

void sort(string, N)
     char *string;
     int N;
{
  int i, j;
  char *min, first;

  min = (char *) malloc(N);
  memcpy(min, string, N);
  
  for(i=0; i<N; i++) {
    first = string[0];
    for(j=0; j<N-1; j++)
      string[j] = string[j+1];
    string[N-1] = first;
    for(j=0; j<N-1; j++)
      if (string[j] != min[j]) {
	if (string[j] < min[j])
	  memcpy(min, string, N);
	break;
      }
  }
  
  memcpy(string, min, N);
  free(min);
}
