#include "xpinball.h"

char poincare_type;

char *colour[NO_COL] = {"white", "red", "pink", "orange", "yellow", "lawngreen",
		   "green", "darkgreen", "cyan", "blue", "indigo", "BlueViolet", 
		   "magenta", "gray90", "gray75", "gray70", 
		    "gray50", "gray45", "gray40", "black"};

void map_cb(), Previous_cb(), Clear(), P_Expose_cb(), P_init(), Calc_dim();
void Draw_map(), Draw_key(), Label_axes(), option_cb(), change_poincare_type();
void convert(), Kill_dialog_cb();
int count_col();
Widget GetTopShell();

void Poincare_cb(w, data, call_data)
     Widget w;
     Data *data;
     XtPointer call_data;
{
  Widget pb, main, commands, option_menu;
  XmString mode, line, arc;
  XSetWindowAttributes attrs;

  if (data->poincare)
    return;
  data->poincare++;
  data->dialog = XtVaCreatePopupShell ("Poincare",
        xmDialogShellWidgetClass, GetTopShell (w),
        XmNdeleteResponse, XmDESTROY,
        NULL);

  main = XtVaCreateWidget ("main",
			xmPanedWindowWidgetClass, data->dialog,
                        XmNsashWidth, 1,
                        XmNsashHeight, 1,
			NULL);
  
  commands = XtVaCreateManagedWidget ("commands",
		      xmRowColumnWidgetClass, main,
		      XmNorientation, XmHORIZONTAL,
		      NULL);
  
  /* add a pushbutton the user can use to clear the canvas */
  pb = XtVaCreateManagedWidget ("Cancel",
				xmPushButtonGadgetClass, commands,
				NULL);
  
  /* if activated, call same cb as XmNinputCallback. */
  XtAddCallback (pb, XmNactivateCallback, Kill_dialog_cb,
		 (XtPointer) data);

  
  pb = XtVaCreateManagedWidget ("Previous",
		xmPushButtonGadgetClass, commands,
		NULL);

  XtAddCallback (pb, XmNactivateCallback, Previous_cb, (XtPointer) data);


  mode = XmStringCreateSimple ("Mode:");
  line = XmStringCreateSimple ("Line");
  arc = XmStringCreateSimple ("Arc");

  option_menu = XmVaCreateSimpleOptionMenu (commands, "option_menu",
        mode, NULL, 0 /*P_initial menu selection*/, option_cb,
        XmVaPUSHBUTTON, line, NULL, NULL, NULL,
        XmVaPUSHBUTTON, arc, NULL, NULL, NULL,
	NULL);

  XmStringFree (mode);
  XmStringFree (line);
  XmStringFree (arc);
  XtManageChild (option_menu);
  XtManageChild(commands);

  data->map = XtVaCreateWidget ("map",
	xmDrawingAreaWidgetClass, main,
	XtVaTypedArg, XmNbackground, XmRString, "white", strlen("white")+1,
	XmNheight, 300,
	XmNwidth, 400,
        NULL);

  XtAddCallback (data->map, XmNinputCallback, map_cb,
		 (XtPointer) data);
  XtAddCallback (data->map, XmNexposeCallback, P_Expose_cb,
		 (XtPointer) data);

  XtManageChild (data->map);
  XtManageChild(main);

  P_init(data);
  attrs.bit_gravity = ForgetGravity;
  XChangeWindowAttributes (XtDisplay(data->map), XtWindow(data->map), 
			   CWBitGravity, &attrs);

  XtPopup (data->dialog, XtGrabNone);
}

void P_init(data)
     Data *data;
{
  int i;
  XGCValues gcv;

  data->rect = NULL;
  data->P_col = NULL;
  data->square = 1;
  data->n_range = 0;
  data->poincare_type = poincare_type = 'l';
  change_poincare_type(data);

  gcv.foreground = BlackPixelOfScreen(XtScreen(data->map));
  data->gc = XCreateGC(XtDisplay (data->map),
      RootWindowOfScreen(XtScreen (data->map)), GCForeground, &gcv);
  if ((data->font_info=XLoadQueryFont(XtDisplay (data->map), "9x15"))==NULL) {
    fprintf(stderr, "font not found\n");
    exit(3);
  }
  XSetFont(XtDisplay (data->map), data->gc, data->font_info->fid);
  XtVaSetValues (data->map, XmNuserData, data->gc, NULL);
  for(i=0; i<NO_COL; i++) {
    data->gc_map[i] = XCreateGC (XtDisplay(data->map),
	   RootWindowOfScreen(XtScreen(data->map)), GCForeground, &gcv);
    SetColour(data->map, data->gc_map[i], colour[i]); 
  }
}

void map_cb(widget, data, call_data)
     Widget widget;
     Data *data;
     XtPointer call_data;
{
  static Position i1, j1;
  Position i2, j2, max, zero;
  double tmp, x1, x2, y1, y2;
  XmDrawingAreaCallbackStruct *cbs = 
        (XmDrawingAreaCallbackStruct *) call_data;
  XEvent *event = cbs->event;
  Range *r;

  if (cbs->reason == XmCR_INPUT) {
    /* activated by DrawingArea input event -- draw lines.
     * Button Down events anchor the P_initial point and Button
     * Up draws from the anchor point to the button-up point.
     */
    if (event->xany.type == ButtonPress) {
      if (event->xbutton.button==Button1) {
	r = &data->range[data->n_range];
	tmp = 1.0/data->size.x;
	zero = OFFSETY1 + data->no_sq.y - 1;
	x1 = r->bl.x + (r->tr.x - r->bl.x)*(event->xbutton.x-OFFSETX1)*tmp;
	y1 = r->bl.y + (r->tr.y - r->bl.y)*(zero-event->xbutton.y)*tmp;
	convert(x1, y1, data);
	DrawRays(data);
       }
      if (event->xbutton.button==Button2) {
	i1 = event->xbutton.x;
	j1 = event->xbutton.y;
      }
    } else if ((event->xany.type == ButtonRelease) &&
	       (event->xbutton.button==Button2)) {
      i2 = event->xbutton.x;
      j2 = event->xbutton.y;
      if (i1==i2 || j1==j2)
	return;
      max = (i1>i2)? i1:i2;
      i1 = (i1<i2)? i1:i2;
      i2 = max;
      max = (j1>j2)? j1:j2;
      j1 = (j1<j2)? j1:j2;
      j2 = max;
      r = &data->range[data->n_range];
      tmp = 1.0/data->size.x;
      x1 = r->bl.x + (r->tr.x - r->bl.x)*(i1-OFFSETX1)*tmp;
      x2 = r->bl.x + (r->tr.x - r->bl.x)*(i2-OFFSETX1)*tmp;
      tmp = 1.0/data->size.y;
      zero = OFFSETY1 + data->no_sq.y - 1;
      y2 = r->bl.y + (r->tr.y - r->bl.y)*(zero-j1)*tmp;
      y1 = r->bl.y + (r->tr.y - r->bl.y)*(zero-j2)*tmp;
      if (x1>data->limit_range.tr.x || x2<data->limit_range.bl.x ||
	  y1>data->limit_range.tr.y || y2<data->limit_range.bl.y)
	return;
      x2 = (x2<data->limit_range.tr.x)? x2:data->limit_range.tr.x;
      x1 = (x1>data->limit_range.bl.x)? x1:data->limit_range.bl.x;
      y2 = (y2<data->limit_range.tr.y)? y2:data->limit_range.tr.y;
      y1 = (y1>data->limit_range.bl.y)? y1:data->limit_range.bl.y;
      data->n_range++;
      r = &data->range[data->n_range];
      r->bl.x=x1;
      r->tr.x=x2;
      r->bl.y=y1;
      r->tr.y=y2;
	
      XDrawRectangle(XtDisplay(data->map), XtWindow(data->map), 
		 data->gc, i1, j1, i2-i1, j2-j1);

      XFlush(XtDisplay(data->map));
      Draw_map(data);
    }
  }
}


void Previous_cb(widget, data, call_data)
     Widget widget;
     Data *data;
     XtPointer call_data;
{
  if(data->n_range)
    data->n_range--;
  Draw_map(data);
}

void P_Expose_cb(widget, data, call_data)
     Widget widget;
     Data *data;
     XtPointer call_data;
{
  if (data->poincare_type!=poincare_type)
    change_poincare_type(data);
  Calc_dim(data);
  Draw_map(data);
}

void Calc_dim(data)
     Data *data;
{
  int i;
  Dimension w, h;

  if (data->P_col!=NULL) {
    for(i=0; i<data->size.x; i++)
      free(data->P_col[i]);
    free(data->P_col);
  }
  if (data->rect!=NULL)
    XFree(data->rect);

  XtVaGetValues(data->map, XmNheight, &h, XmNwidth, &w, NULL);
  setv(data->no_sq, (w-OFFSETX1-OFFSETX2)/data->square,
       (h-OFFSETY1-OFFSETY2)/data->square);
  smulv(data->size, data->square, data->no_sq);

  data->P_col = (char **) calloc(data->no_sq.x, sizeof(char*));
  for(i=0; i<data->no_sq.x; i++)
    data->P_col[i] = (char *) malloc(data->no_sq.y);
  data->rect = (XRectangle *) XtCalloc(data->no_sq.x*data->no_sq.y, sizeof(XRectangle));
  for(i=0; i<data->no_sq.x*data->no_sq.y; i++) {
    data->rect[i].width = data->square+1;
    data->rect[i].height = data->square+1;
  }
}

void Draw_map(data)
     Data *data;
{
  int c, max, min, zero;
  int i, i0, j, j0, Ncol[NO_COL], Pt[NO_COL], pt;
  double x, x_step, y, y_step;
  Range *r;

  r = &data->range[data->n_range];

  x_step=(r->tr.x-r->bl.x)/((double) data->no_sq.x);
  y_step=(r->tr.y-r->bl.y)/((double) data->no_sq.y);

  for(i=0; i<NO_COL; i++)
    Ncol[i] = 0;
  min = 1000;
  max = 0;
  for(i=0, x=r->bl.x; i<data->no_sq.x; i++, x+=x_step)
    for(j=0, y=r->bl.y; j<data->no_sq.y; j++, y+=y_step) {
      convert(x, y, data);
      c = (char) count_col(data);
      max = (c>max)? c:max;
      min = (c<min)? c:min;
      c = (c>NO_COL-1)? NO_COL-1:c;
      data->P_col[i][j] = c;
      Ncol[c]++;
    }
  Pt[0] = 0;
  for(i=1; i<NO_COL; i++)
    Pt[i] = Pt[i-1] + Ncol[i-1];

  if (max>NO_COL)
    fprintf(stderr, "Not enough colours max trajectory = %d p_collisions\n", max);
    
  XClearWindow(XtDisplay(data->map), XtWindow(data->map));
  XDrawRectangle(XtDisplay(data->map), XtWindow(data->map), 
		 data->gc, OFFSETX1-1, OFFSETY1-1, data->size.x+2, data->size.y+2);
  Draw_key(data);
  Label_axes(data);

  zero = OFFSETY1 + data->no_sq.y - 1;
  for(i=0, i0=OFFSETX1; i<data->no_sq.x; i++, i0+=data->square)
    for(j=0, j0=zero; j<data->no_sq.y; j++, j0-=data->square) {
      pt = Pt[data->P_col[i][j]]++;
      data->rect[pt].x = i0;
      data->rect[pt].y = j0;
    }

  XFillRectangles( XtDisplay(data->map), XtWindow(data->map), 
		  data->gc_map[0], data->rect, Ncol[0]);
  for(c=0; c<NO_COL; c++) {
    if(Ncol[c])
      XFillRectangles( XtDisplay(data->map), XtWindow(data->map), 
		      data->gc_map[c], data->rect+Pt[c-1], Ncol[c]);
  }
}

void Label_axes(data)
     Data *data;
{
  char buf[50];
  int len, width, n, tmp;

  tmp = OFFSETY1+data->size.y+20;
  n = data->n_range;
  sprintf(buf, "%g", data->range[n].bl.x);
  len = strlen(buf);
  width = XTextWidth(data->font_info, buf, len);
  XDrawString(XtDisplay(data->map), XtWindow(data->map), 
	      data->gc, OFFSETX1-width/2, tmp, buf, len);
  sprintf(buf, "%g", data->range[n].tr.x);
  len = strlen(buf);
  width = XTextWidth(data->font_info, buf, len);
  XDrawString(XtDisplay(data->map), XtWindow(data->map), 
	      data->gc, OFFSETX1+data->size.x-width/2, tmp, buf, len);
  sprintf(buf, "%g", data->range[n].bl.y);
  len = strlen(buf);
  width = XTextWidth(data->font_info, buf, len);
  XDrawString(XtDisplay(data->map), XtWindow(data->map), 
	      data->gc, OFFSETX1-width-5, OFFSETY1+data->size.y+5, buf, len);
  sprintf(buf, "%g", data->range[n].tr.y);
  len = strlen(buf);
  width = XTextWidth(data->font_info, buf, len);
  XDrawString(XtDisplay(data->map), XtWindow(data->map), 
	      data->gc, OFFSETX1-width-5, OFFSETY1+5, buf, len);
}

void Draw_key(data)
     Data *data;
{
  char buf[10];
  int i, h, offset, ht, len;

  h = (data->size.y-OFFSETY1-OFFSETY2-20)/NO_COL;
  ht = NO_COL*h;
  offset = OFFSETY1 + (data->size.y-ht)/2;    
  for(i=0; i<NO_COL; i++)
    XFillRectangle( XtDisplay(data->map), XtWindow(data->map), 
		      data->gc_map[i], OFFSETX1/2, offset+i*h, 10, h);
  XDrawRectangle(XtDisplay(data->map), XtWindow(data->map), 
		      data->gc, OFFSETX1/2-1, offset-1, 11, ht+1);
  XDrawString(XtDisplay(data->map), XtWindow(data->map), 
	      data->gc, OFFSETX1/2+14, offset+9, "0", 2);
  sprintf(buf, "%d", NO_COL-1);
  len = strlen(buf);
  XDrawString(XtDisplay(data->map), XtWindow(data->map), 
	      data->gc, OFFSETX1/2+14, offset+ht, buf, len);
}

void convert(x, y, data)
  double x, y;
  Data *data;
{
  double tmp;
  Vector R, Rt;

  if (data->poincare_type=='l') {
    setv(data->init_pos, 0, -x);
    tmp = sqrt(1.0-y*y);
    setv(data->init_vel, y, tmp);
  } else {
    tmp = sin(y);
    setv(R, sqrt(1-tmp*tmp), tmp);
    setv(Rt, -R.y, R.x);
    tmp = sqrt(1-x*x);
    smulv(Rt, x, Rt);
    smulv(data->init_vel, tmp, R);
    addv(data->init_vel, data->init_vel, Rt);
    smulv(data->init_pos, data->a, R);
    addv(data->init_pos, data->init_pos, data->disk[0]);
    smulv(R, 1.0e-6, data->init_vel);
    addv(data->init_pos, data->init_pos, R);
  }
}

int count_col(data)
     Data *data;
{
  int len, i, cd, d_tmp, no_hits;
  Vector pos, vel, dd, v_tmp, perp;
  double tmp1, tmp2, t_tmp, time;

  no_hits = 0;
  cd = NO_HIT;
  pos = data->init_pos;
  vel = data->init_vel;

  for (len=0; len<MAX_HIT; len++) {
    d_tmp = NO_HIT;
    time = INF_T;
    for(i=0; i<ND; i++) {
      if (i!=cd) {
	subv(dd, data->disk[i], pos);
	if ((tmp1 = dotv(vel, dd)) > 0.0) {
	  if ((tmp2=tmp1*tmp1-dotv(dd, dd)+data->asq) > 0.0) {
	    if ((t_tmp=tmp1-sqrt(tmp2))<time) {
	      d_tmp = i;
	      time = t_tmp;
	    }
	  }
	}
      }
    }
    if (time<INF_T) {
      no_hits++;
      cd = d_tmp;
      smulv(v_tmp, time, vel);
      addv(pos, pos, v_tmp);
      subv(perp, pos, data->disk[cd]);
      smulv(perp, data->ia, perp);
      tmp1 = 2.0*dotv(vel, perp);
      smulv(v_tmp, tmp1, perp);
      subv(vel, vel, v_tmp);
    } else
      break;
  }
  return(no_hits);
}

void option_cb(menu_item, client_data, call_data)
     Widget menu_item;
     XtPointer client_data;
     XtPointer call_data;
{
  long item_no = (long) client_data;

  if (item_no == 0)
    poincare_type = 'l';
  else
    poincare_type = 'a';
}

void change_poincare_type(data)
     Data *data;
{
  data->poincare_type = poincare_type;

  if (poincare_type=='l') {
    setv(data->range[0].bl, 0.0, -1.0);
    setv(data->range[0].tr, LEN, 1.0);
    setvv(data->limit_range.bl, data->range[0].bl);
    setvv(data->limit_range.tr, data->range[0].tr);
  } else {
    setv(data->range[0].bl, -1.0, -0.5*PI);
    setv(data->range[0].tr, 1.0, 0.5*PI);
    setvv(data->limit_range.bl, data->range[0].bl);
    setvv(data->limit_range.tr, data->range[0].tr);
  }
  data->n_range = 0;
}

void Kill_dialog_cb(w, data, call_data)
     Widget w;
     Data *data;
     XtPointer call_data;
{
  data->poincare = 0;
  XtDestroyWidget(data->dialog);
}

Widget GetTopShell(w)
     Widget w;
{
  while (w && !XtIsWMShell (w))
    w = XtParent (w);
  return w;
}
